from __future__ import annotations

import contextlib
import sqlite3
import sys
from collections.abc import Generator
from collections.abc import ItemsView
from collections.abc import Iterable
from collections.abc import Iterator
from collections.abc import Mapping
from functools import wraps
from typing import Callable
from typing import Generic
from typing import TypeVar

if sys.version_info >= (3, 10):  # pragma >=3.10 cover
    from typing import ParamSpec
else:  # pragma <3.10 cover
    from typing_extensions import ParamSpec


@contextlib.contextmanager
def connect(db_path: str) -> Generator[sqlite3.Connection, None, None]:
    """Context manager to connect to a sqlite database.

    :param db_path: path to the sqlite database

    :return: A Generator yielding an open sqlite connection
    """
    with contextlib.closing(
        sqlite3.connect(
            db_path,
            detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES,
        ),
    ) as db:
        db.row_factory = sqlite3.Row
        with db:
            yield db


P = ParamSpec('P')
R = TypeVar('R')


def retry(
        retries: int,
        exceptions: tuple[type[Exception], ...] = (Exception,),
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """Decorator to retry a function ``retries`` times when a specific
    exceptions is raised (defined in ``exceptions``). If any other exception is
    raised, it will not retry the function.

    It can be used like this: A function is decorated and it is retried a
    maximum of 10 times when a ``ValueError`` or ``KeyError`` is raised. Every
    other exception will instantly be raised.

    .. code-block:: python

        @retry(retries=10, exceptions=(ValueError, KeyError))
        def my_func():
            ...

    :param retries: number of times a function is retried
    :param exceptions: the exceptions to except and retry
    """
    def retry_dec(f: Callable[P, R]) -> Callable[P, R]:
        @wraps(f)
        def inner(*args: P.args, **kwargs: P.kwargs) -> R:
            curr_tries = 0
            while True:
                try:
                    return f(*args, **kwargs)
                except exceptions:
                    if curr_tries >= retries:
                        raise
                    curr_tries += 1

        return inner
    return retry_dec


K = TypeVar('K')
V = TypeVar('V')


class FrozenDict(Generic[K, V]):
    """Immutable, generic implementation of a frozen dictionary."""

    def __init__(self, d: Mapping[K, V]) -> None:
        self._d = d

    def __getitem__(self, k: K) -> V:
        return self._d[k]

    def __contains__(self, k: K) -> bool:
        return k in self._d

    def __iter__(self) -> Iterator[K]:
        yield from self._d

    def get(self, k: K) -> V | None:
        return self._d.get(k)

    def values(self) -> Iterable[V]:
        return self._d.values()

    def keys(self) -> Iterable[K]:
        return self._d.keys()

    def items(self) -> ItemsView[K, V]:
        return self._d.items()

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self._d})'



from __future__ import annotations

from collections.abc import Generator
from contextlib import contextmanager
from datetime import datetime
from datetime import timezone
from typing import Any
from typing import Literal
from typing import NamedTuple

import serial

from vpf_730.utils import connect
from vpf_730.utils import FrozenDict

"""
Frozen Dictionary mapping the precipitation types abbreviations to their full
name form.
"""
PRECIP_TYPES = FrozenDict({
    'NP': 'No precipitation',
    'DZ-': 'Slight drizzle',
    'DZ': 'Moderate drizzle',
    'DZ+': 'Heavy drizzle',
    'RA-': 'Slight rain',
    'RA': 'Moderate rain',
    'RA+': 'Heavy rain',
    'SN-': 'Slight snow',
    'SN': 'Moderate snow',
    'SN+': 'Heavy snow',
    'UP': 'Indeterminate precipitation type',
    'GS': 'Small Hail',
    'GR': 'Hail',
    'X': 'Initial value or error',
})

"""
Frozen Dictionary mapping the obstruction to vision abbreviations to their full
name form.
"""
OBSTRUCTION_TO_VISION = FrozenDict({
    '': 'No obstruction',
    'HZ': 'Haze',
    'FG': 'Fog',
    'DU': 'Dust',
    'FU': 'Smoke',
    'BR': 'Mist',
})


MEASUREMENT_TABLE = '''\
        CREATE TABLE IF NOT EXISTS measurements(
            timestamp INT PRIMARY KEY,
            sensor_id INT NOT NULL,
            last_measurement_period INT,
            time_since_report INT,
            optical_range NUMERIC,
            precipitation_type_msg TEXT,
            obstruction_to_vision TEXT,
            receiver_bg_illumination NUMERIC,
            water_in_precip NUMERIC,
            temp NUMERIC,
            nr_precip_particles INT,
            transmission_eq NUMERIC,
            exco_less_precip_particle NUMERIC,
            backscatter_exco NUMERIC,
            self_test VARCHAR(3),
            total_exco NUMERIC
        )
    '''


class Measurement(NamedTuple):
    """``NamedTuple`` class representing a Measurement from the VPF-730 sensor.
    Data as defined in the manual:
    https://www.biral.com/wp-content/uploads/2019/07/VPF-710-730-750-Manual-102186.08E.pdf

    :param timestamp: Timestamp in milliseconds (UTC)
    :param sensor_id: Sensor identification number set by the user
    :param last_measurement_period: Last measurement period in seconds
    :param time_since_report: Time since this report was generated seconds
    :param optical_range: Meteorological optical range in km
    :param precipitation_type_msg: Precipitation type message (one of:  :const:`PRECIP_TYPES`)
    :param obstruction_to_vision: Obstruction to vision message (one of: :const:`OBSTRUCTION_TO_VISION`)
    :param receiver_bg_illumination: Receiver background illumination
    :param water_in_precip: Amount of water in precipitation in last measurement period in mm
    :param temp: Temperature in °C
    :param nr_precip_particles: Number of precipitation particles detected in last measurement period
    :param transmission_eq: Transmissometer equivalent EXCO km :superscript:`-1`
    :param exco_less_precip_particle: EXCO less precipitation particle component km :superscript:`-1`
    :param backscatter_exco: Backscatter EXCO km :superscript:`-1`
    :param self_test: Self-Test and Monitoring (see Manual section 4.2)
    :param total_exco: Total EXCO km :superscript:`-1`
    """  # noqa: E501
    timestamp: int
    sensor_id: int
    last_measurement_period: int
    time_since_report: int
    optical_range: float
    precipitation_type_msg: str
    obstruction_to_vision: str
    receiver_bg_illumination: float
    water_in_precip: float
    temp: float
    nr_precip_particles: int
    transmission_eq: float
    exco_less_precip_particle: float
    backscatter_exco: float
    self_test: str
    total_exco: float

    @property
    def precipitation_type_msg_readable(self) -> str:
        """Return the precipitation type message as a human readable message
        instead of the 2 digit abbreviation. This is for convenience and is
        just using :const:`PRECIP_TYPES` internally.

        :return: text message containing the precipitation type
        """
        return PRECIP_TYPES[self.precipitation_type_msg]

    @property
    def obstruction_to_vision_readable(self) -> str:
        """Return the obstruction to vision type message as a human readable
        message instead of the 2 digit abbreviation.T his is just for
        convenience and is using :const:`OBSTRUCTION_TO_VISION` internally.

        :return: text message containing the obstruction to vision type
        """
        return OBSTRUCTION_TO_VISION[self.obstruction_to_vision]

    @classmethod
    def from_msg(cls, msg: bytes) -> Measurement:
        """Constructs a new :func:`Measurement` from the bytes read.

        :param msg: bytes representing a message read from the sensor using
            :func:`VPF730.measure` e.g.
            ``b'PW01,0060,0000,001.19 KM,NP ,HZ,00.06,00.0000,+020.5 C,0000,002.51,002.51,+011.10,  0000,000,OOO,002.51'``

        :return: a new instance of :func:`Measurement`.
        """  # noqa: E501
        # checksum is off by default
        msg_str = msg.decode()
        msg_list = msg_str.strip().split(',')
        # checks
        precipitation_type_msg = msg_list[4].strip()
        if precipitation_type_msg not in PRECIP_TYPES:
            raise ValueError(
                f'unknown precipitation type {precipitation_type_msg!r}. '
                f'Must be one of: {", ".join(PRECIP_TYPES)}',
            )

        obstruction_to_vision = msg_list[5].strip()
        if obstruction_to_vision not in OBSTRUCTION_TO_VISION:
            raise ValueError(
                f'unknown obstruction to vision type {obstruction_to_vision!r}'
                f'. Must be one of: {", ".join(obstruction_to_vision)}',
            )

        return cls(
            timestamp=int(datetime.now(timezone.utc).timestamp()),
            # strip the message header
            sensor_id=int(msg_list[0].lstrip('PW')),
            last_measurement_period=int(msg_list[1]),
            time_since_report=int(msg_list[2]),
            optical_range=float(msg_list[3].rstrip('KM')),
            precipitation_type_msg=precipitation_type_msg,
            obstruction_to_vision=obstruction_to_vision,
            receiver_bg_illumination=float(msg_list[6]),
            water_in_precip=float(msg_list[7]),
            temp=float(msg_list[8].rstrip('C')),
            nr_precip_particles=int(msg_list[9]),
            transmission_eq=float(msg_list[10]),
            # TODO: convert this status
            exco_less_precip_particle=float(msg_list[11]),
            backscatter_exco=float(msg_list[12]),
            self_test=msg_list[15],
            total_exco=float(msg_list[16]),
        )

    def to_csv(self, fname: str | None = None, sep: str = ',') -> str | None:
        """Convert a measurement to a csv formatted string in this format:
        ``timestamp,sensor_id,last_measurement_period,time_since_report,
        optical_range,precipitation_type_msg,obstruction_to_vision,
        receiver_bg_illumination,water_in_precip,temp,nr_precip_particles,
        transmission_eq,exco_less_precip_particle,backscatter_exco,
        self_test,total_exco``.

        if ``fname`` is set, write it to a file (using append mode).

        :param fname: optional - a filename to write the data to
        :param sep: separator to use for the csv, default: ``,``

        :return: a string containing the measurements formatted as csv or
            ``None`` when written to a file.
        """
        csv_str = sep.join(str(i) for i in self)
        if fname is not None:
            with open(fname, 'a') as f:
                # a brand new file, write the header
                if f.tell() == 0:
                    f.write(f'{self.csv_header()}\n')
                f.write(f'{csv_str}\n')
            return None
        else:
            return csv_str

    def csv_header(self, sep: str = ',') -> str:
        """Create a csv-file header containing all fields in the correct order:
        ``timestamp,sensor_id,last_measurement_period,time_since_report,
        optical_range,precipitation_type_msg,obstruction_to_vision,
        receiver_bg_illumination,water_in_precip,temp,nr_precip_particles,
        transmission_eq,exco_less_precip_particle,backscatter_exco,
        self_test,total_exco``

        :param sep: separator to use for the csv, default: ``,``

        :return: a string containing the csv header
        """
        return sep.join(self._fields)

    def to_db(self, db_path: str) -> None:
        """Insert the measurement into a sqlite database

        :param db_path: path to the sqlite database
        """
        with connect(db_path) as db:
            db.execute(MEASUREMENT_TABLE)
            db.execute(
                '''\
                INSERT INTO measurements(
                    timestamp,
                    sensor_id,
                    last_measurement_period,
                    time_since_report,
                    optical_range,
                    precipitation_type_msg,
                    obstruction_to_vision,
                    receiver_bg_illumination,
                    water_in_precip,
                    temp,
                    nr_precip_particles,
                    transmission_eq,
                    exco_less_precip_particle,
                    backscatter_exco,
                    self_test,
                    total_exco
                )
                VALUES (
                    :timestamp,
                    :sensor_id,
                    :last_measurement_period,
                    :time_since_report,
                    :optical_range,
                    :precipitation_type_msg,
                    :obstruction_to_vision,
                    :receiver_bg_illumination,
                    :water_in_precip,
                    :temp,
                    :nr_precip_particles,
                    :transmission_eq,
                    :exco_less_precip_particle,
                    :backscatter_exco,
                    :self_test,
                    :total_exco
                )
                ''',
                self._asdict(),
            )


class VPF730:
    """A class for interacting with the VPF-730 sensor. Please also see the
    pySerial documentation: https://pyserial.readthedocs.io/

    :param port: serial port the VPF-730 sensor is connected to
    :param baudrate: Baud rate such as 9600 or 115200 etc
    :param bytesize: Number of data bits. Possible values ``5``, ``6``, ``7``, ``8``
    :param parity: Enable parity checking. Possible values: ``N``, ``E``, ``O``, ``M``, ``S``
    :param stopbits: Number of stop bits. Possible values: ``1``, ``1.5``, ``2``
    :param timeout: Set a read timeout value in seconds
    :param xonxoff: Enable software flow control
    :param rtscts:  Enable hardware (RTS/CTS) flow control
    :param write_timeout: Set a write timeout value in seconds
    :param dsrdtr: Enable hardware (DSR/DTR) flow control
    :param inter_byte_timeout: Inter-character timeout, None to disable (default)
    :param exclusive: Set exclusive access mode (POSIX only). A port cannot be
        opened in exclusive access mode if it is already open in exclusive
        access mode.
    :param kwargs: any additional keyword arguments
    """  # noqa: E501

    def __init__(
            self,
            port: str,
            *,
            baudrate: int = 1200,
            bytesize: Literal[5, 6, 7, 8] = 8,
            parity: Literal['N', 'E', 'O', 'M', 'S'] = 'N',
            # TODO: can this take an enum?
            stopbits: int = 1,
            timeout: float = 3,
            xonxoff: bool = False,
            rtscts: bool = False,
            write_timeout: float | None = None,
            dsrdtr: bool = False,
            inter_byte_timeout: float | None = None,
            exclusive: bool | None = None,
            **kwargs: Any,
    ) -> None:
        self.port = port
        self.baudrate = baudrate
        self.bytesize = bytesize
        self.parity = parity
        self.stopbits = stopbits
        self.timeout = timeout
        self.xonxoff = xonxoff
        self.rtscts = rtscts
        self.write_timeout = write_timeout
        self.dsrdtr = dsrdtr
        self.inter_byte_timeout = inter_byte_timeout
        self.exclusive = exclusive
        self._kwargs = kwargs

        # defer opening
        self._ser = serial.Serial()
        self._ser.port = self.port
        self._ser.baudrate = self.baudrate
        self._ser.bytesize = self.bytesize
        self._ser.parity = self.parity
        self._ser.stopbits = self.stopbits
        self._ser.timeout = self.timeout
        self._ser.xonxoff = self.xonxoff
        self._ser.rtscts = self.rtscts
        self._ser.write_timeout = self.write_timeout
        self._ser.dsrdtr = self.dsrdtr
        self._ser.inter_byte_timeout = self.inter_byte_timeout
        self._ser.exclusive = self.exclusive

    @contextmanager
    def open_ser(self) -> Generator[None, None, None]:
        """Context manager for opening and closing the serial port"""
        try:
            self._ser.open()
            yield
        finally:
            self._ser.close()

    def send_command(self, command: str) -> bytes:
        """Send an ASCII command to the VPF-730. A detailed description can be
        found in the Biral VPF-XXX Manual starting on page 59:
        https://www.biral.com/wp-content/uploads/2019/07/VPF-710-730-750-Manual-102186.08E.pdf

        :param command: A valid command e.g: ``D?``

        :return: the response of the sensor as bytest
        """
        with self.open_ser():
            cmd = f'{command}\r\n'
            self._ser.write(cmd.encode())
            return self._ser.read_until(b'\r\n')

    def measure(self, polled_mode: bool = True) -> Measurement | None:
        """Read the VPF-730 sensor using the previously configured serial
        interface and return a :func:`Measurement` or None, if the sensor did
        not return any data.

        :param polled_mode: read the sensor in polled mode. The mode can be set
            in the sensor using the ``OSAMx`` command, where ``x`` is ``0`` for
            automatic message transmission disabled and ``1`` for automatic
            message transmission enabled (default: ``True``).

        :return: a new :func:`Measurement` containing the data read from the
            sensor
        """
        with self.open_ser():
            if polled_mode is True:
                self._ser.write(b'D?\r\n')

            msg = self._ser.read_until(b'\r\n')
            if msg:
                return Measurement.from_msg(msg)
            else:
                return None



from .logger import Logger
from .logger import LoggerConfig
from .sender import Sender
from .sender import SenderConfig
from .vpf_730 import Measurement
from .vpf_730 import OBSTRUCTION_TO_VISION
from .vpf_730 import PRECIP_TYPES
from .vpf_730 import VPF730


__all__ = [
    'Logger', 'LoggerConfig', 'Sender', 'SenderConfig', 'Measurement',
    'OBSTRUCTION_TO_VISION', 'PRECIP_TYPES', 'VPF730',
]



from __future__ import annotations

import argparse
import configparser
import logging
import os
import sys
from argparse import RawDescriptionHelpFormatter
from collections.abc import Sequence

from vpf_730.logger import Logger
from vpf_730.logger import LoggerConfig
from vpf_730.sender import Sender
from vpf_730.sender import SenderConfig
from vpf_730.vpf_730 import VPF730

# VPF730_SENTRY_DSN and VPF730_SENTRY_SAMPLE_RATE env var need to be set for
# monitoring
try:
    import sentry_sdk
    sample_rate = int(os.environ.get('VPF730_SENTRY_SAMPLE_RATE', 0))
    sentry_sdk.init(
        dsn=os.environ.get('VPF730_SENTRY_DSN'),
        traces_sample_rate=sample_rate,
    )
except ImportError:  # pragma: no cover
    pass

loglevel = os.environ.get('VPF730_LOGLEVEL', logging.ERROR)
logger = logging.getLogger(__name__)
logger.setLevel(loglevel)
hdlr = logging.StreamHandler(sys.stdout)
logger.addHandler(hdlr)


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        'vpf-730',
        formatter_class=RawDescriptionHelpFormatter,
    )
    subparsers = parser.add_subparsers(
        dest='command',
        help='Chose the type of process to run',
    )
    # set up the parser for manual communication
    comm_parser = subparsers.add_parser(
        'comm',
        help='Manually send ASCII commands to the VPF-730 sensor',
    )
    comm_parser.add_argument(
        'ascii_command',
        help=(
            'The ASCII command to be send to the VPF-730 sensor e.g: D?. The '
            'carriage return (\\r) and line feed (\\n) at the end MUST NOT be '
            'part of the command. It is appended automatically.'
        ),
    )
    comm_parser.add_argument(
        '--serial-port',
        help='Serial port the VPF-730 sensor is connected to, e.g /dev/ttyS0',
    )
    comm_file_config = comm_parser.add_argument_group('config from file')
    comm_file_config.description = (
        'Reads the configuration from a file and overrides all previous CLI '
        'options'
    )
    comm_file_config.add_argument(
        '-c', '--config',
        help='Path to an .ini config file',
    )
    comm_parser.epilog = (
        'If no argument for --serial-port is provided, the configuration will '
        'be read from the environment variable VPF730_PORT.'
    )

    # set up the parser for the logger
    logger_parser = subparsers.add_parser(
        'logger',
        help=(
            'Run a logger taking measurements from the VPF-730 sensor and '
            'store them in a local database'
        ),
        formatter_class=RawDescriptionHelpFormatter,
    )
    logger_cli_config = logger_parser.add_argument_group('config from CLI')
    logger_cli_config.add_argument(
        '--local-db',
        default='vpf_730_local.db',
        help='Path to the local database',
    )
    logger_cli_config.add_argument(
        '--serial-port',
        help='Serial port the VPF-730 sensor is connected to, e.g /dev/ttyS0',
    )
    logger_cli_config.add_argument(
        '--log-interval',
        help=(
            'the interval to be used for logging e.g. 1 for every minute '
            '(minimum), 30 for 30 minutes (maximum)'
        ),
        default=1,
        metavar='[1-30]',
        type=int,
    )
    file_config = logger_parser.add_argument_group('config from file')
    file_config.description = (
        'Reads the configuration from a file and overrides all previous CLI '
        'options'
    )
    file_config.add_argument(
        '-c', '--config',
        help='Path to an .ini config file',
    )
    logger_parser.epilog = (
        'If no arguments are provided, the configuration will be read from '
        'the environment variables.\n'
        '  - VPF730_LOCAL_DB\n'
        '  - VPF730_PORT\n'
        '  - VPF730_LOG_INTERVAL\n'
        'For variable descriptions see the CLI arguments above'
    )

    # set up the parser for the sender
    sender_parser = subparsers.add_parser(
        'sender',
        help=(
            'Synchronize the local database with a remote endpoint by sending '
            'data via a POST request'
        ),
        formatter_class=RawDescriptionHelpFormatter,
    )
    sender_parser.add_argument(
        '--local-db',
        default='vpf_730_local.db',
        help='Path to the local database',
    )
    sender_parser.add_argument(
        '--send-interval',
        help=(
            'The interval in which data should be send to the remote server '
            '1, every minute (minimum), 30 for 30 minutes (maximum)'
        ),
        metavar='[1-30]',
        type=int,
        default=5,
    )
    sender_parser.add_argument(
        '--get-endpoint',
        help=(
            'API endpoint to get the status of the remote server i.e. what is '
            'the latest data e.g. https://api.example/com/vpf-730/status. The '
            'API-Key must be provided as an environment variable '
            'VPF730_API_KEY=mykey'
        ),
    )
    sender_parser.add_argument(
        '--post-endpoint',
        help=(
            'API endpoint to send the data to e.g. '
            'https://api.example/com/vpf-730/data. The API-Key must be '
            'provided as an environment variable VPF730_API_KEY=mykey'
        ),
    )
    sender_parser.add_argument(
        '--max-req-len',
        help=(
            'the maximum number of measurements that are allowed to be send '
            'in a single request'
        ),
        default=512,
        type=int,
    )
    file_config = sender_parser.add_argument_group('config from file')
    file_config.description = (
        'Reads the configuration from a file and overrides all previous CLI '
        'options'
    )
    file_config.add_argument(
        '-c', '--config',
        help='Path to an .ini config file',
    )
    sender_parser.epilog = (
        'If no arguments are provided, the configuration will be read from '
        'the environment variables.\n'
        '  - VPF730_LOCAL_DB\n'
        '  - VPF730_SEND_INTERVAL\n'
        '  - VPF730_GET_ENDPOINT\n'
        '  - VPF730_POST_ENDPOINT\n'
        '  - VPF730_MAX_REQ_LEN\n'
        '  - VPF730_API_KEY\n'
        'For variable descriptions see the CLI arguments above'
    )
    return parser


def main(argv: Sequence[str] | None = None) -> int:
    parser = build_parser()
    args = parser.parse_args(argv)
    if args.command == 'logger':
        if args.config:
            logger_cfg = LoggerConfig.from_file(path=args.config)
        elif args.serial_port:
            logger_cfg = LoggerConfig.from_argparse(args=args)
        else:
            logger_cfg = LoggerConfig.from_env()

        vpf_logger = Logger(cfg=logger_cfg)
        try:
            logger.info('starting logger with configuration: %s', logger_cfg)
            vpf_logger.run()
        except KeyboardInterrupt:
            logger.info('logger received shutdown signal...')
            vpf_logger.logging = False
            return 0

    elif args.command == 'sender':
        if args.config:
            sender_cfg = SenderConfig.from_file(path=args.config)
        elif args.get_endpoint or args.post_endpoint:
            if args.get_endpoint is None:
                raise parser.error(
                    'must set --get-endpoint when configuring via CLI',
                )
            if args.post_endpoint is None:
                raise parser.error(
                    'must set --post-endpoint when configuring via CLI',
                )
            sender_cfg = SenderConfig.from_argparse(args=args)
        else:
            sender_cfg = SenderConfig.from_env()

        sender = Sender(cfg=sender_cfg)
        try:
            logger.info('starting sender with configuration: %s', sender_cfg)
            sender.run()
        except KeyboardInterrupt:
            logger.info('sender received shutdown signal...')
            sender.sending = False
            return 0
    elif args.command == 'comm':
        if args.config:
            config = configparser.ConfigParser()
            config.read(args.config)
            serial_port = config['vpf_730']['serial_port']
        elif args.serial_port:
            serial_port = args.serial_port
        else:
            serial_port = os.environ['VPF730_PORT']

        com_sender = VPF730(port=serial_port)
        ret = com_sender.send_command(args.ascii_command)
        print(ret.decode())
    else:
        raise NotImplementedError()

    return 0


if __name__ == '__main__':
    raise SystemExit(main())



from __future__ import annotations

import argparse
import configparser
import json
import logging
import os
import time
import urllib.error
import urllib.request
from datetime import datetime
from datetime import timezone
from typing import NamedTuple
from typing import TypedDict

from vpf_730.utils import connect

logger = logging.getLogger(__name__)


class SenderConfig(NamedTuple):
    """A class representing the configuration of sender.

    :param local_db: path to the local sqlite database, where the measurements
        are stored
    :param send_interval: interval in minutes to send data to the endpoint
        minimum 1, maximum 30 (every 30 minutes)
    :param max_req_len: maximum number of measurements to send in one request
    :param get_endpoint: http endpoint to get the status from (latest data)
    :param post_endpoint: http endpoint where the data should be posted to
    """
    local_db: str
    send_interval: int
    get_endpoint: str
    post_endpoint: str
    max_req_len: int
    api_key: str

    @classmethod
    def from_env(cls) -> SenderConfig:
        """Constructs a new :func:`LoggerConfig` from environment variables.

        * ``VPF730_LOCAL_DB`` - path to the sqlite database which is used to store data locally
        * ``VPF730_SEND_INTERVAL`` - interval in minutes to send data to the endpoint
        * ``VPF730_GET_ENDPOINT`` - http endpoint to get the status from (latest data)
        * ``VPF730_POST_ENDPOINT`` - http endpoint where the data should be posted to
        * ``VPF730_API_KEY`` - the API-key used to authenticate when sending the requests in

        :return: a new instance of :func:`SenderConfig` created from
            environment variables.
        """  # noqa: E501
        return cls(
            local_db=os.environ['VPF730_LOCAL_DB'],
            send_interval=int(os.environ['VPF730_SEND_INTERVAL']),
            max_req_len=int(os.environ['VPF730_MAX_REQ_LEN']),
            get_endpoint=os.environ['VPF730_GET_ENDPOINT'],
            post_endpoint=os.environ['VPF730_POST_ENDPOINT'],
            api_key=os.environ['VPF730_API_KEY'],
        )

    @classmethod
    def from_file(cls, path: str) -> SenderConfig:
        """Constructs a new :func:`SenderConfig` from a provided ``.ini``
        config file with this format:

            .. code-block:: ini

                [vpf_730]
                local_db=local.db
                send_interval=5
                get_endpoint=https://api.example/com/vpf-730/status
                post_endpoint=https://api.example/com/vpf-730/data
                max_req_len=512
                api_key=deadbeef


        :param path: path to the ``.ini`` config file with the structure above

        :return: a new instance of :func:`LoggerConfig` created from a config
            file
        """
        config = configparser.ConfigParser()
        config.read(path)
        return cls(
            config['vpf_730']['local_db'],
            int(config['vpf_730']['send_interval']),
            config['vpf_730']['get_endpoint'],
            config['vpf_730']['post_endpoint'],
            int(config['vpf_730']['max_req_len']),
            config['vpf_730']['api_key'],
        )

    @classmethod
    def from_argparse(cls, args: argparse.Namespace) -> SenderConfig:
        """Constructs a new :func:`SenderConfig` from a
        :func:`argparse.Namespace`, created by the argument parser returned by
        :func:`vpf_730.main.build_parser`.

        :param args: arguments returned from the argument parser created by
            :func:`vpf_730.main.build_parser`

        :return: a new instance of :func:`SenderConfig` created from CLI
            arguments
        """
        return cls(
            local_db=args.local_db,
            send_interval=args.send_interval,
            get_endpoint=args.get_endpoint,
            post_endpoint=args.post_endpoint,
            max_req_len=args.max_req_len,
            api_key=os.environ['VPF730_API_KEY'],
        )

    def __repr__(self) -> str:
        return (
            f'{type(self).__name__}('
            f'local_db={self.local_db!r}, '
            f'send_interval={self.send_interval!r}, '
            f'get_endpoint={self.get_endpoint!r}, '
            f'post_endpoint={self.post_endpoint!r}, '
            f'max_req_len={self.max_req_len!r}, '
            f'api_key=***)'
        )


class MeasurementDict(TypedDict):
    timestamp: int
    sensor_id: int
    last_measurement_period: int
    time_since_report: int
    optical_range: float
    precipitation_type_msg: str
    obstruction_to_vision: str
    receiver_bg_illumination: float
    water_in_precip: float
    temp: float
    nr_precip_particles: int
    transmission_eq: float
    exco_less_precip_particle: float
    backscatter_exco: float
    self_test: str
    total_exco: float


class Sender:
    def __init__(self, cfg: SenderConfig) -> None:
        self.cfg = cfg
        self.sending = True

    @property
    def _sending(self) -> bool:
        # this is a hack for being able to test this
        return self.sending  # pragma: no cover

    def run(self) -> None:
        prev_minute = -1
        while self._sending is True:
            time.sleep(.1)
            now = datetime.now(timezone.utc)
            # don't accidentally log the same timestamp twice
            if (
                now.minute % self.cfg.send_interval == 0 and
                now.second == 0 and
                now.minute != prev_minute
            ):
                last_date = self.get_remote_timestamp()
                data = self.get_data_from_db(start=last_date)
                if not data:
                    continue

                # we don't exceed the max_req_len and can send the data as one
                if len(data) <= self.cfg.max_req_len:
                    self.post_data_to_remote(data=data)
                    continue

                counter = 0
                post_stack: list[MeasurementDict] = []
                for row in data:
                    post_stack.append(row)
                    if counter == self.cfg.max_req_len - 1:
                        self.post_data_to_remote(data=post_stack)
                        post_stack = []
                        counter = 0
                    else:
                        counter += 1

                # if we have data left which does not fill a max size request
                else:
                    if post_stack:
                        self.post_data_to_remote(data=post_stack)

    def get_remote_timestamp(self) -> int:
        status_req = urllib.request.Request(
            url=self.cfg.get_endpoint,
            headers={
                'Authorization': self.cfg.api_key,
                'Content-type': 'application/json',
            },
        )
        try:
            status_resp = urllib.request.urlopen(status_req)
            status_resp_str = status_resp.read().decode()
            return json.loads(status_resp_str)['latest_date']
        except urllib.error.HTTPError as e:
            msg = json.loads(e.read().decode())
            logger.exception('http error getting latest date: %s', msg)
            raise

    def post_data_to_remote(self, data: list[MeasurementDict]) -> None:
        post_data = json.dumps({'data': data}).encode()
        req = urllib.request.Request(
            url=self.cfg.post_endpoint,
            data=post_data,
            headers={
                'Authorization': self.cfg.api_key,
                'Content-type': 'application/json',
            },
        )
        try:
            urllib.request.urlopen(req)
        except urllib.error.HTTPError as e:
            msg = json.loads(e.read().decode())
            logger.exception('http error sending date: %s', msg)
            raise

    def get_data_from_db(self, start: int) -> list[MeasurementDict]:
        """Get data from the db starting after ``start``

        :param start: unix timestamp (UTC) after which to get data

        :return: data
        """
        with connect(self.cfg.local_db) as db:
            query = '''\
                SELECT
                    timestamp,
                    sensor_id,
                    last_measurement_period,
                    time_since_report,
                    optical_range,
                    precipitation_type_msg,
                    obstruction_to_vision,
                    receiver_bg_illumination,
                    water_in_precip,
                    temp,
                    nr_precip_particles,
                    transmission_eq,
                    exco_less_precip_particle,
                    backscatter_exco,
                    self_test,
                    total_exco
                FROM measurements
                WHERE timestamp > ?
                ORDER BY timestamp
            '''
            ret = db.execute(query, (start,))
            val = ret.fetchall()
            # https://github.com/python/mypy/issues/8890
            md = MeasurementDict
            return [md(i) for i in val]  # type: ignore[call-arg, misc]



from __future__ import annotations

import argparse
import configparser
import os
import time
from datetime import datetime
from datetime import timezone
from typing import NamedTuple

from vpf_730.vpf_730 import VPF730


class LoggerError(Exception):
    """Base class for errors raised by the logger"""
    pass


class LoggerConfigError(LoggerError):
    """Exception that is raised when the configuration is invalid"""
    pass


class LoggerConfig(NamedTuple):
    """A class representing the configuration of logger.

    :param local_db: path to the local sqlite database, where the measurements
        are stored
    :param serial_port: serial port that the VPF-730 sensor is connected to
    :param log_interval: the log interval in minutes (between 0 and 30)
    """
    local_db: str
    serial_port: str
    log_interval: int

    @classmethod
    def from_env(cls) -> LoggerConfig:
        """Constructs a new :func:`LoggerConfig` from environment variables.

        * ``VPF730_LOCAL_DB`` - path to the sqlite database which is used as  queue
        * ``VPF730_PORT`` - serial port that the VPF-730 sensor is connected to
        * ``VPF730_LOG_INTERVAL`` - interval used for logging e.g. 1 for every minute

        :return: a new instance of :func:`LoggerConfig` created from
            environment variables.
        """  # noqa: E501
        return cls(
            local_db=os.environ['VPF730_LOCAL_DB'],
            serial_port=os.environ['VPF730_PORT'],
            log_interval=int(os.environ['VPF730_LOG_INTERVAL']),
        )

    @classmethod
    def from_file(cls, path: str) -> LoggerConfig:
        """Constructs a new :func:`LoggerConfig` from a provided ``.ini``
        config file with this format:

            .. code-block:: ini

                [vpf_730]
                local_db=local.db
                serial_port=/dev/ttyS0
                log_interval=1

        :param path: path to the ``.ini`` config file with the structure above

        :return: a new instance of :func:`LoggerConfig` created from a config
            file
        """
        config = configparser.ConfigParser()
        config.read(path)
        return cls(
            config['vpf_730']['local_db'],
            config['vpf_730']['serial_port'],
            int(config['vpf_730']['log_interval']),
        )

    @classmethod
    def from_argparse(cls, args: argparse.Namespace) -> LoggerConfig:
        """Constructs a new :func:`LoggerConfig` from an
        :func:`argparse.Namespace`, created by the argument parser returned by
        :func:`vpf_730.main.build_parser`.

        :param args: arguments returned from the argument parser created by
            :func:`vpf_730.main.build_parser`

        :return: a new instance of :func:`LoggerConfig` created from CLI
            arguments
        """
        if args.log_interval is None or not (0 < args.log_interval <= 30):
            raise LoggerConfigError(
                'the log interval must be set between 1 and 30',
            )
        return cls(
            local_db=args.local_db,
            serial_port=args.serial_port,
            log_interval=args.log_interval,
        )


class Logger:
    def __init__(self, cfg: LoggerConfig) -> None:
        self.cfg = cfg
        self.logging = True
        self.vpf_730 = VPF730(port=cfg.serial_port)

    @property
    def _logging(self) -> bool:
        # this is a hack for being able to test this
        return self.logging  # pragma: no cover

    def run(self) -> None:
        prev_minute = -1
        while self._logging is True:
            time.sleep(.1)
            now = datetime.now(timezone.utc)
            # don't accidentally log the same timestamp twice
            if (
                    now.minute % self.cfg.log_interval == 0 and
                    now.second == 0 and
                    now.minute != prev_minute
            ):
                measurement = self.vpf_730.measure()
                if measurement is not None:  # pragma: no branch
                    measurement.to_db(db_path=self.cfg.local_db)

                prev_minute = now.minute
